# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta
from abc import abstractmethod
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

from pydantic import validate_arguments
from pydantic.dataclasses import dataclass

from kolena._api.v1.generic import TestRun as API
from kolena._utils.datatypes import get_args
from kolena._utils.datatypes import get_origin
from kolena._utils.validators import ValidatorConfig
from kolena.workflow import GroundTruth
from kolena.workflow import Inference
from kolena.workflow import TestCase
from kolena.workflow import TestSample
from kolena.workflow import TestSuite
from kolena.workflow._datatypes import DataObject
from kolena.workflow._validators import get_data_object_field_types
from kolena.workflow._validators import validate_data_object_type
from kolena.workflow._validators import validate_scalar_data_object_type

# include for backwards compatibility -- definitions moved to ./plot.py
# noreorder
from kolena.workflow import AxisConfig  # noqa: F401
from kolena.workflow import Plot  # noqa: F401
from kolena.workflow import Curve  # noqa: F401
from kolena.workflow import CurvePlot  # noqa: F401
from kolena.workflow import Histogram  # noqa: F401
from kolena.workflow import BarPlot  # noqa: F401
from kolena.workflow import ConfusionMatrix  # noqa: F401


@dataclass(frozen=True, config=ValidatorConfig)
class MetricsTestSample(DataObject, metaclass=ABCMeta):
    """
    Test-sample-level metrics produced by an [`Evaluator`][kolena.workflow.Evaluator].

    This class should be subclassed with the relevant fields for a given workflow.

    Examples here may include the number of true positive detections on an image, the mean IOU of inferred polygon(s)
    with ground truth polygon(s), etc.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        _validate_metrics_test_sample_type(cls)


@dataclass(frozen=True, config=ValidatorConfig)
class MetricsTestCase(DataObject, metaclass=ABCMeta):
    """
    Test-case-level metrics produced by an [`Evaluator`][kolena.workflow.Evaluator].

    This class should be subclassed with the relevant fields for a given workflow.

    Test-case-level metrics are aggregate metrics like [`precision`][kolena.workflow.metrics.precision],
    [`recall`][kolena.workflow.metrics.recall], and [`f1_score`][kolena.workflow.metrics.f1_score]. Any and all
    aggregate metrics that fit a workflow should be defined here.

    ## Nesting Aggregate Metrics

    `MetricsTestCase` supports nesting metrics objects, for e.g. reporting class-level metrics within a test case that
    contains multiple classes. Example usage:

    ```python
    @dataclass(frozen=True)
    class PerClassMetrics(MetricsTestCase):
        Class: str
        Precision: float
        Recall: float
        F1: float
        AP: float

    @dataclass(frozen=True)
    class TestCaseMetrics(MetricsTestCase):
        macro_Precision: float
        macro_Recall: float
        macro_F1: float
        mAP: float
        PerClass: List[PerClassMetrics]
    ```

    Any `str`-type fields (e.g. `Class` in the above example) will be used as identifiers when displaying nested metrics
    on Kolena. For best results, include at least one `str`-type field in nested metrics definitions.

    When comparing nested metrics from multiple models, an `int`-type column with any of the following names will be
    used for sample size in statistical significance calculations: `N`, `n`, `nTestSamples`, `n_test_samples`,
    `sampleSize`, `sample_size`, `SampleSize`.

    For a detailed overview of this feature, see the [:kolena-diagram-tree-16: Nesting Test Case
    Metrics](../../advanced-usage/nesting-test-case-metrics.md) advanced usage guide.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        _validate_metrics_test_case_type(cls)


@dataclass(frozen=True, config=ValidatorConfig)
class MetricsTestSuite(DataObject, metaclass=ABCMeta):
    """
    Test-suite-level metrics produced by an [`Evaluator`][kolena.workflow.Evaluator].

    This class should be subclassed with the relevant fields for a given workflow.

    Test-suite-level metrics typically measure performance across test cases, e.g. penalizing variance across different
    subsets of a benchmark.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        _validate_metrics_test_suite_type(cls)


@dataclass(frozen=True, config=ValidatorConfig)
class EvaluatorConfiguration(DataObject, metaclass=ABCMeta):
    """
    Configuration for an [`Evaluator`][kolena.workflow.Evaluator].

    Example evaluator configurations may specify:

    - Fixed confidence thresholds at which detections are discarded.
    - Different algorithms/strategies used to compute confidence thresholds
        (e.g. "accuracy optimal" for a classification-type workflow).
    """

    @abstractmethod
    def display_name(self) -> str:
        """
        The name to display for this configuration in Kolena. Must be implemented when extending
        [`EvaluatorConfiguration`][kolena.workflow.EvaluatorConfiguration].
        """
        raise NotImplementedError


class Evaluator(metaclass=ABCMeta):
    """
    An `Evaluator` transforms inferences into metrics.

    Metrics are computed at the individual test sample level ([`MetricsTestSample`][kolena.workflow.MetricsTestSample]),
    in aggregate at the test case level ([`MetricsTestCase`][kolena.workflow.MetricsTestCase]), and across populations
    at the test suite level ([`MetricsTestSuite`][kolena.workflow.MetricsTestSuite]).

    Test-case-level plots ([`Plot`][kolena.workflow.Plot]) may also be computed.

    :param configurations: The configurations at which to perform evaluation. Instance methods such as
        [`compute_test_sample_metrics`][kolena.workflow.Evaluator.compute_test_sample_metrics] are called once per test
        case per configuration.
    """

    configurations: List[EvaluatorConfiguration]
    """The configurations with which to perform evaluation, provided on instantiation."""

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, configurations: Optional[List[EvaluatorConfiguration]] = None):
        if configurations is not None and len(configurations) == 0:
            raise ValueError("empty configuration list provided, at least one configuration required or 'None'")
        self.configurations = configurations or []
        if len({configuration.display_name() for configuration in self.configurations}) < len(self.configurations):
            raise ValueError("all configurations must have distinct display names")

    def display_name(self) -> str:
        """The name to display for this evaluator in Kolena. Defaults to the name of this class."""
        return type(self).__name__

    @abstractmethod
    def compute_test_sample_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> List[Tuple[TestSample, MetricsTestSample]]:
        """
        Compute metrics for every test sample in a test case, i.e. one
        [`MetricsTestSample`][kolena.workflow.MetricsTestSample] object for each of the provided test samples.

        Must be implemented.

        :param test_case: The [`TestCase`][kolena.workflow.TestCase] to which the provided test samples and ground
            truths belong.
        :param inferences: The test samples, ground truths, and inferences for all entries in a test case.
        :param configuration: The evaluator configuration to use. Empty for implementations that are not configured.
        :return: [`TestSample`][kolena.workflow.TestSample]-level metrics for each provided test sample.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_test_case_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[MetricsTestSample],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> MetricsTestCase:
        """
        Compute aggregate metrics ([`MetricsTestCase`][kolena.workflow.MetricsTestCase]) across a test case.

        Must be implemented.

        :param test_case: The test case in question.
        :param inferences: The test samples, ground truths, and inferences for all entries in a test case.
        :param metrics: The [`TestSample`][kolena.workflow.TestSample]-level metrics computed by
            [`compute_test_sample_metrics`][kolena.workflow.Evaluator.compute_test_sample_metrics], provided
            in the same order as `inferences`.
        :param configuration: The evaluator configuration to use. Empty for implementations that are not configured.
        :return: [`TestCase`][kolena.workflow.TestCase]-level metrics for the provided test case.
        """
        raise NotImplementedError

    @validate_arguments(config=ValidatorConfig)
    def compute_test_case_plots(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[MetricsTestSample],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> Optional[List[Plot]]:
        """
        Optionally compute any number of plots to visualize the results for a test case.

        :param test_case: The test case in question
        :param inferences: The test samples, ground truths, and inferences for all entries in a test case.
        :param metrics: The [`TestSample`][kolena.workflow.TestSample]-level metrics computed by
            [`compute_test_sample_metrics`][kolena.workflow.Evaluator.compute_test_sample_metrics], provided
            in the same order as `inferences`.
        :param configuration: the evaluator configuration to use. Empty for implementations that are not configured.
        :return: Zero or more plots for this test case at this configuration.
        """
        return None  # not required

    @validate_arguments(config=ValidatorConfig)
    def compute_test_suite_metrics(
        self,
        test_suite: TestSuite,
        metrics: List[Tuple[TestCase, MetricsTestCase]],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> Optional[MetricsTestSuite]:
        """
        Optionally compute [`TestSuite`][kolena.workflow.TestSuite]-level metrics
        ([`MetricsTestSuite`][kolena.workflow.MetricsTestSuite]) across the provided `test_suite`.

        :param test_suite: The test suite in question.
        :param metrics: The [`TestCase`][kolena.workflow.TestCase]-level metrics computed by
            [`compute_test_case_metrics`][kolena.workflow.Evaluator.compute_test_case_metrics].
        :param configuration: The evaluator configuration to use. Empty for implementations that are not configured.
        :return: The [`TestSuite`][kolena.workflow.TestSuite]-level metrics for this test suite.
        """
        return None  # not required


def _validate_metrics_test_sample_type(metrics_test_sample_type: Type[MetricsTestSample]) -> None:
    # TODO: support special structure for ImagePair test sample types?
    validate_data_object_type(metrics_test_sample_type)


def _validate_metrics_test_case_type(metrics_test_case_type: Type[DataObject]) -> None:
    validate_scalar_data_object_type(metrics_test_case_type, supported_list_types=[MetricsTestCase])

    # validate that there is only one level of nesting
    for field_name, field_type in get_data_object_field_types(metrics_test_case_type).items():
        origin = get_origin(field_type)
        if origin is not list:  # only need to check lists, as MetricsTestCase is only allowed in lists
            continue
        # expand e.g. List[Union[MetricsA, MetricsB]] into [MetricsA, MetricsB]
        list_arg_types = [t for arg_type in get_args(field_type) for t in get_args(arg_type) or [arg_type]]
        for arg_type in list_arg_types:
            if arg_type is None:
                raise ValueError(f"Unsupported optional metrics object in field '{field_name}'")
            try:
                validate_scalar_data_object_type(arg_type)
            except ValueError:
                raise ValueError(f"Unsupported doubly-nested metrics object in field '{field_name}'")


def _validate_metrics_test_suite_type(metrics_test_suite_type: Type[DataObject]) -> None:
    validate_scalar_data_object_type(metrics_test_suite_type)


def _maybe_evaluator_configuration_to_api(
    configuration: Optional[EvaluatorConfiguration],
) -> Optional[API.EvaluatorConfiguration]:
    if configuration is None:
        return None
    return API.EvaluatorConfiguration(display_name=configuration.display_name(), configuration=configuration._to_dict())


def _maybe_display_name(configuration: Optional[EvaluatorConfiguration]) -> Optional[str]:
    if configuration is None:
        return None
    return configuration.display_name()


@validate_arguments(config=ValidatorConfig)
def _configuration_description(configuration: Optional[EvaluatorConfiguration]) -> str:
    display_name = _maybe_display_name(configuration)
    return f"(configuration: {display_name})" if display_name is not None else ""



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def accuracy(true_positives: int, false_positives: int, false_negatives: int, true_negatives: int) -> float:
    """
    Accuracy represents the proportion of inferences that are correct (including both positives and negatives).

    $$
    \\text{Accuracy} = \\frac{\\text{# TP} + \\text{# TN}}
    {\\text{# TP} + \\text{# FP} + \\text{# FN} + \\text{# TN}}
    $$

    <div class="grid cards" markdown>
    - :kolena-metrics-glossary-16: Metrics Glossary: [Accuracy ↗](../../metrics/accuracy.md)
    </div>

    :param true_positives: Number of true positive inferences.
    :param false_positives: Number of false positive inferences.
    :param false_negatives: Number of false negatives.
    :param true_negatives: Number of true negatives.
    """
    numerator = true_positives + true_negatives
    denominator = true_positives + false_positives + false_negatives + true_negatives
    return numerator / denominator if denominator > 0 else 0


def precision(true_positives: int, false_positives: int) -> float:
    """
    Precision represents the proportion of inferences that are correct.

    $$
    \\text{Precision} = \\frac{\\text{# True Positives}}{\\text{# True Positives} + \\text{# False Positives}}
    $$

    <div class="grid cards" markdown>
    - :kolena-metrics-glossary-16: Metrics Glossary: [Precision ↗](../../metrics/precision.md)
    </div>

    :param true_positives: Number of true positive inferences.
    :param false_positives: Number of false positive inferences.
    """
    denominator = true_positives + false_positives
    return true_positives / denominator if denominator > 0 else 0


def recall(true_positives: int, false_negatives: int) -> float:
    """
    Recall represents the proportion of ground truths that were successfully predicted.

    $$
    \\text{Recall} = \\frac{\\text{# True Positives}}{\\text{# True Positives} + \\text{# False Negatives}}
    $$

    <div class="grid cards" markdown>
    - :kolena-metrics-glossary-16: Metrics Glossary: [Recall ↗](../../metrics/recall.md)
    </div>

    :param true_positives: Number of true positive inferences.
    :param false_negatives: Number of false negatives.
    """
    denominator = true_positives + false_negatives
    return true_positives / denominator if denominator > 0 else 0


def f1_score(true_positives: int, false_positives: int, false_negatives: int) -> float:
    """
    F<sub>1</sub>-score is the harmonic mean between [`precision`][kolena.workflow.metrics.precision] and
    [`recall`][kolena.workflow.metrics.recall].

    $$
    \\begin{align}
    \\text{F1}
    &= \\frac{2}{\\frac{1}{\\text{Precision}} + \\frac{1}{\\text{Recall}}} \\\\[1em]
    &= 2 \\times \\frac{\\text{Precision} \\times \\text{Recall}}{\\text{Precision} + \\text{Recall}}
    \\end{align}
    $$

    <div class="grid cards" markdown>
    - :kolena-metrics-glossary-16: Metrics Glossary: [F<sub>1</sub>-score ↗](../../metrics/f1-score.md)
    </div>

    :param true_positives: Number of true positive inferences.
    :param false_positives: Number of false positive inferences.
    :param false_negatives: Number of false negatives.
    """
    prec = precision(true_positives, false_positives)
    rec = recall(true_positives, false_negatives)
    denominator = prec + rec
    return 2 * prec * rec / denominator if denominator > 0 else 0



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
import pickle
from base64 import b64encode
from typing import List
from typing import Tuple

import numpy as np
import pandas as pd
from dacite import from_dict

from kolena._api.v1.generic import Search as API
from kolena._experimental.search._internal.datatypes import LocatorEmbeddingsDataFrameSchema
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame
from kolena._utils.consts import BatchSize
from kolena._utils.dataframes.validators import validate_df_schema
from kolena.errors import InputValidationError


def upload_embeddings(key: str, embeddings: List[Tuple[str, np.ndarray]]) -> None:
    """
    Upload a list of search embeddings corresponding to sample locators.

    :param key: String value uniquely corresponding to the model used to extract the embedding vectors.
        This is typically a locator.
    :param embeddings: List of locator-embedding pairs, as tuples. Locators should be string values, while embeddings
        should be an `numpy.typing.ArrayLike` of numeric values.
    :raises InputValidationError: The provided embeddings input is not of a valid format
    """
    init_response = init_upload()
    locators, search_embeddings = [], []
    for locator, embedding in embeddings:
        if not np.issubdtype(embedding.dtype, np.number):
            raise InputValidationError("unexepected non-numeric embedding dtype")
        locators.append(locator)
        search_embeddings.append(b64encode(pickle.dumps(embedding.astype(np.float32))).decode("utf-8"))
    df_embeddings = pd.DataFrame(dict(key=[key] * len(embeddings), locator=locators, embedding=search_embeddings))
    df_validated = validate_df_schema(df_embeddings, LocatorEmbeddingsDataFrameSchema)

    log.info(f"uploading embeddings for key '{key}'")
    upload_data_frame(df=df_validated, batch_size=BatchSize.UPLOAD_EMBEDDINGS.value, load_uuid=init_response.uuid)
    request = API.UploadEmbeddingsRequest(
        uuid=init_response.uuid,
    )
    res = krequests.post(
        endpoint_path=API.Path.EMBEDDINGS.value,
        data=json.dumps(dataclasses.asdict(request)),
    )
    krequests.raise_for_status(res)
    data = from_dict(data_class=API.UploadEmbeddingsResponse, data=res.json())
    log.success(f"uploaded embeddings for key '{key}' on {data.n_samples} samples")



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Annotations are visualized in Kolena as overlays on top of [`TestSample`][kolena.workflow.TestSample] objects.

The following annotation types are available:

- [`BoundingBox`][kolena.workflow.annotation.BoundingBox]
- [`Polygon`][kolena.workflow.annotation.Polygon]
- [`Polyline`][kolena.workflow.annotation.Polyline]
- [`Keypoints`][kolena.workflow.annotation.Keypoints]
- [`BoundingBox3D`][kolena.workflow.annotation.BoundingBox3D]
- [`SegmentationMask`][kolena.workflow.annotation.SegmentationMask]
- [`BitmapMask`][kolena.workflow.annotation.BitmapMask]
- [`ClassificationLabel`][kolena.workflow.annotation.ClassificationLabel]

For example, when viewing images in the Studio, any annotations (such as lists of
[`BoundingBox`][kolena.workflow.annotation.BoundingBox] objects) present in the
[`TestSample`][kolena.workflow.TestSample], [`GroundTruth`][kolena.workflow.GroundTruth],
[`Inference`][kolena.workflow.Inference], or [`MetricsTestSample`][kolena.workflow.MetricsTestSample] objects are
rendered on top of the image.
"""
import dataclasses
from abc import ABCMeta
from functools import reduce
from typing import Dict
from typing import List
from typing import Tuple

from pydantic.dataclasses import dataclass

from kolena._utils.validators import ValidatorConfig
from kolena.workflow._datatypes import DataType
from kolena.workflow._datatypes import TypedDataObject


class _AnnotationType(DataType):
    BOUNDING_BOX = "BOUNDING_BOX"
    POLYGON = "POLYGON"
    POLYLINE = "POLYLINE"
    KEYPOINTS = "KEYPOINTS"
    BOUNDING_BOX_3D = "BOUNDING_BOX_3D"
    SEGMENTATION_MASK = "SEGMENTATION_MASK"
    BITMAP_MASK = "BITMAP_MASK"
    CLASSIFICATION_LABEL = "LABEL"

    @staticmethod
    def _data_category() -> str:
        return "ANNOTATION"


@dataclass(frozen=True, config=ValidatorConfig)
class Annotation(TypedDataObject[_AnnotationType], metaclass=ABCMeta):
    """The base class for all annotation types."""


@dataclass(frozen=True, config=ValidatorConfig)
class BoundingBox(Annotation):
    """
    Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices.

    The reserved fields `width`, `height`, `area`, and `aspect_ratio` are automatically populated with values derived
    from the provided coordinates.
    """

    top_left: Tuple[float, float]
    """The top left vertex (in `(x, y)` pixel coordinates) of this bounding box."""

    bottom_right: Tuple[float, float]
    """The bottom right vertex (in `(x, y)` pixel coordinates) of this bounding box."""

    width: float = dataclasses.field(init=False)
    height: float = dataclasses.field(init=False)
    area: float = dataclasses.field(init=False)
    aspect_ratio: float = dataclasses.field(init=False)

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.BOUNDING_BOX

    def __post_init__(self) -> None:
        object.__setattr__(self, "width", self.bottom_right[0] - self.top_left[0])
        object.__setattr__(self, "height", self.bottom_right[1] - self.top_left[1])
        object.__setattr__(self, "area", self.width * self.height)
        object.__setattr__(self, "aspect_ratio", self.width / self.height if self.height != 0 else 0)


@dataclass(frozen=True, config=ValidatorConfig)
class LabeledBoundingBox(BoundingBox):
    """
    Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices and a string
    label.
    """

    label: str
    """The label (e.g. model classification) associated with this bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredBoundingBox(BoundingBox):
    """
    Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices and a float
    score.
    """

    score: float
    """The score (e.g. model confidence) associated with this bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabeledBoundingBox(BoundingBox):
    """
    Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices, a string
    label, and a float score.
    """

    label: str
    """The label (e.g. model classification) associated with this bounding box."""

    score: float
    """The score (e.g. model confidence) associated with this bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class Polygon(Annotation):
    """Arbitrary polygon specified by three or more pixel coordinates."""

    points: List[Tuple[float, float]]
    """The sequence of `(x, y)` pixel coordinates comprising the boundary of this polygon."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.POLYGON

    def __post_init__(self) -> None:
        if len(self.points) < 3:
            raise ValueError(f"{type(self).__name__} must have at least three points ({len(self.points)} provided)")


@dataclass(frozen=True, config=ValidatorConfig)
class LabeledPolygon(Polygon):
    """Arbitrary polygon specified by three or more pixel coordinates and a string label."""

    label: str
    """The label (e.g. model classification) associated with this polygon."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredPolygon(Polygon):
    """
    Arbitrary polygon specified by three or more pixel coordinates and a float score.
    """

    score: float
    """The score (e.g. model confidence) associated with this polygon."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabeledPolygon(Polygon):
    """
    Arbitrary polygon specified by three or more pixel coordinates with a string label and a float score.
    """

    label: str
    """The label (e.g. model classification) associated with this polygon."""

    score: float
    """The score (e.g. model confidence) associated with this polygon."""


@dataclass(frozen=True, config=ValidatorConfig)
class Keypoints(Annotation):
    """Array of any number of keypoints specified in pixel coordinates."""

    points: List[Tuple[float, float]]
    """The sequence of discrete `(x, y)` pixel coordinates comprising this keypoints annotation."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.KEYPOINTS


@dataclass(frozen=True, config=ValidatorConfig)
class Polyline(Annotation):
    """Polyline with any number of vertices specified in pixel coordinates."""

    points: List[Tuple[float, float]]
    """The sequence of connected `(x, y)` pixel coordinates comprising this polyline."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.POLYLINE


@dataclass(frozen=True, config=ValidatorConfig)
class BoundingBox3D(Annotation):
    """
    Three-dimensional cuboid bounding box in a right-handed coordinate system.

    Specified by `(x, y, z)` coordinates for the `center` of the cuboid, `(x, y, z)` `dimensions`, and a `rotation`
    parameter specifying the degrees of rotation about each axis `(x, y, z)` ranging `[-π, π]`.

    The reserved field `volume` is automatically derived from the provided `dimensions`.
    """

    center: Tuple[float, float, float]
    """`(x, y, z)` coordinates specifying the center of the bounding box."""

    dimensions: Tuple[float, float, float]
    """`(x, y, z)` measurements specifying the dimensions of the bounding box."""

    rotations: Tuple[float, float, float]
    """Rotations in degrees about each `(x, y, z)` axis."""

    volume: float = dataclasses.field(init=False)

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.BOUNDING_BOX_3D

    def __post_init__(self) -> None:
        object.__setattr__(self, "volume", reduce(lambda a, b: a * b, self.dimensions))


@dataclass(frozen=True, config=ValidatorConfig)
class LabeledBoundingBox3D(BoundingBox3D):
    """[`BoundingBox3D`][kolena.workflow.annotation.BoundingBox3D] with an additional string label."""

    label: str
    """The label associated with this 3D bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredBoundingBox3D(BoundingBox3D):
    """[`BoundingBox3D`][kolena.workflow.annotation.BoundingBox3D] with an additional float score."""

    score: float
    """The score associated with this 3D bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabeledBoundingBox3D(BoundingBox3D):
    """[`BoundingBox3D`][kolena.workflow.annotation.BoundingBox3D] with an additional string label and float score."""

    label: str
    """The label associated with this 3D bounding box."""

    score: float
    """The score associated with this 3D bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class SegmentationMask(Annotation):
    """
    Raster segmentation mask. The `locator` is the URL to the image file representing the segmentation mask.

    The segmentation mask must be rendered as a single-channel, 8-bit-depth (grayscale) image. For the best results,
    use a lossless file format such as PNG. Each pixel's value is the numerical ID of its class label, as specified in
    the `labels` map. Any pixel value not present in the `labels` map is rendered as part of the background.

    For example, `labels = {255: "object"}` will highlight all pixels with the value of 255 as `"object"`. Every
    other pixel value will be transparent.
    """

    labels: Dict[int, str]
    """Mapping of unique label IDs (pixel values) to unique label values."""

    locator: str
    """URL of the segmentation mask image."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.SEGMENTATION_MASK


@dataclass(frozen=True, config=ValidatorConfig)
class BitmapMask(Annotation):
    """Arbitrary bitmap mask. The `locator` is the URL to the image file representing the mask."""

    locator: str
    """URL of the bitmap data."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.BITMAP_MASK


@dataclass(frozen=True, config=ValidatorConfig)
class ClassificationLabel(Annotation):
    """Label of classification."""

    label: str
    """String label for this classification."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.CLASSIFICATION_LABEL


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredClassificationLabel(ClassificationLabel):
    """Classification label with accompanying score."""

    score: float
    """Score associated with this label."""


_ANNOTATION_TYPES = [
    BoundingBox,
    LabeledBoundingBox,
    ScoredBoundingBox,
    ScoredLabeledBoundingBox,
    Polygon,
    LabeledPolygon,
    ScoredPolygon,
    ScoredLabeledPolygon,
    Keypoints,
    Polyline,
    BoundingBox3D,
    LabeledBoundingBox3D,
    ScoredBoundingBox3D,
    ScoredLabeledBoundingBox3D,
    SegmentationMask,
    BitmapMask,
    ClassificationLabel,
    ScoredClassificationLabel,
]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import sys
from typing import Any
from typing import Iterator
from typing import Optional
from typing import TypeVar

import termcolor
from tqdm.auto import tqdm

from kolena._utils.instrumentation import log_telemetry
from kolena._utils.state import _client_state


LOGGER_NAME = "kolena"
_logger = logging.getLogger(LOGGER_NAME)
_logger.addHandler(logging.NullHandler())


def _colored(message: str, color: Optional[str] = None) -> str:
    # disable coloration on windows ('nt') as it does not work without hassle
    if color is not None and os.name != "nt" and color in termcolor.COLORS:
        return termcolor.colored(message, color=color)
    else:  # color printing not available on this system
        return message


def _print(message: str, color: Optional[str] = None, **kwargs: Any) -> None:
    print(_colored(message, color), **kwargs)


def _print_header(**kwargs: Any) -> None:
    _print(f"{LOGGER_NAME}> ", color="magenta", end="", **kwargs)


def success(message: str, **kwargs: Any) -> None:
    if _client_state.verbose:
        _print_header()
        _print(message, **kwargs, color="green")
    _logger.info(message, **kwargs)


def info(message: str, **kwargs: Any) -> None:
    if _client_state.verbose:
        _print_header()
        _print(message, **kwargs)
    _logger.info(message, **kwargs)


def warn(message: str, **kwargs: Any) -> None:
    if _client_state.verbose:
        _print_header(file=sys.stderr)
        _print(message, **kwargs, color="yellow", file=sys.stderr)
    _logger.warning(message, **kwargs)


def error(message: str, exception: Optional[BaseException], **kwargs: Any) -> None:
    _print_header(file=sys.stderr)
    _print(message, **kwargs, color="red", file=sys.stderr)
    _logger.error(message, exc_info=exception, **kwargs)

    if exception and _client_state.telemetry:
        log_telemetry(exception)


T = TypeVar("T")


def progress_bar(iterator: Iterator[T], desc: Optional[str] = None, **kwargs: Any) -> Iterator[T]:
    if _client_state.verbose:
        desc_base = "kolena> " if is_notebook else _colored("kolena> ", color="magenta")
        desc_full = f"{desc_base}{desc}" if desc is not None else desc_base
        iterator = tqdm(iterator, desc=desc_full, **kwargs)
    yield from iterator


# Logic copied from https://stackoverflow.com/a/39662359
def _is_notebook() -> bool:
    if "IPython" not in sys.modules:
        return False
    try:
        get_ipython = sys.modules["IPython"].get_ipython
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


is_notebook = _is_notebook()



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Simplified interface for [`Evaluator`][kolena.workflow.Evaluator] implementations.
"""
import dataclasses
import json
from abc import ABCMeta
from abc import abstractmethod
from dataclasses import field
from inspect import signature
from typing import Callable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

from pydantic.dataclasses import dataclass

from kolena._api.v1.generic import TestRun as API
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.state import is_client_initialized
from kolena._utils.validators import ValidatorConfig
from kolena.workflow import EvaluatorConfiguration
from kolena.workflow import GroundTruth as BaseGroundTruth
from kolena.workflow import Inference as BaseInference
from kolena.workflow import MetricsTestCase
from kolena.workflow import MetricsTestSample as BaseMetricsTestSample
from kolena.workflow import MetricsTestSuite
from kolena.workflow import Plot
from kolena.workflow import TestCase
from kolena.workflow import TestSample as BaseTestSample
from kolena.workflow.evaluator import _configuration_description

TestSample = TypeVar("TestSample", bound=BaseTestSample)
GroundTruth = TypeVar("GroundTruth", bound=BaseGroundTruth)
Inference = TypeVar("Inference", bound=BaseInference)
MetricsTestSample = TypeVar("MetricsTestSample", bound=BaseMetricsTestSample)


class TestCases(metaclass=ABCMeta):
    """
    Provides an iterator method for grouping test-sample-level metric results with the test cases that they belong to.
    """

    @abstractmethod
    def iter(
        self,
        test_samples: List[TestSample],
        ground_truths: List[GroundTruth],
        inferences: List[Inference],
        metrics_test_sample: List[MetricsTestSample],
    ) -> Iterator[Tuple[TestCase, List[TestSample], List[GroundTruth], List[Inference], List[MetricsTestSample]]]:
        """
        Matches test sample metrics to the corresponding test cases that they belong to.

        :param test_samples: All unique test samples within the test run, sequenced in the same order as the other
            parameters.
        :param ground_truths: Ground truths corresponding to `test_samples`, sequenced in the same order.
        :param inferences: Inferences corresponding to `test_samples`, sequenced in the same order.
        :param metrics_test_sample: Test-sample-level metrics corresponding to `test_samples`, sequenced in the
            same order.
        :return: Iterator that groups each test case in the test run to the lists of member test samples, inferences,
            and test-sample-level metrics.
        """
        raise NotImplementedError


@dataclass(frozen=True, config=ValidatorConfig)
class EvaluationResults:
    """
    A bundle of metrics computed for a test run grouped at the test-sample-level, test-case-level, and test-suite-level.
    Optionally includes [`Plot`s][kolena.workflow.Plot] at the test-case-level.
    """

    metrics_test_sample: List[Tuple[BaseTestSample, BaseMetricsTestSample]]
    """
    Sample-level metrics, extending [`MetricsTestSample`][kolena.workflow.MetricsTestSample], for every provided test
    sample.
    """

    metrics_test_case: List[Tuple[TestCase, MetricsTestCase]]
    """
    Aggregate metrics, extending [`MetricsTestCase`][kolena.workflow.MetricsTestCase], computed across each test case
    yielded from [`TestCases.iter`][kolena.workflow.TestCases.iter].
    """

    plots_test_case: List[Tuple[TestCase, List[Plot]]] = field(default_factory=list)
    """Optional test-case-level plots."""

    metrics_test_suite: Optional[MetricsTestSuite] = None
    """Optional test-suite-level metrics, extending [`MetricsTestSuite`][kolena.workflow.MetricsTestSuite]."""


ConfiguredEvaluatorFunction = Callable[
    [List[TestSample], List[GroundTruth], List[Inference], TestCases, EvaluatorConfiguration],
    Optional[EvaluationResults],
]
UnconfiguredEvaluatorFunction = Callable[
    [List[TestSample], List[GroundTruth], List[Inference], TestCases],
    Optional[EvaluationResults],
]
BasicEvaluatorFunction = Union[ConfiguredEvaluatorFunction, UnconfiguredEvaluatorFunction]
"""
`BasicEvaluatorFunction` provides a function-based evaluator interface that takes
the inferences for all test samples in a test suite and a [`TestCases`][kolena.workflow.TestCases] as input and computes
the corresponding test-sample-level, test-case-level, and test-suite-level metrics (and optionally plots) as output.

Example implementation, relying on `compute_per_sample` and `compute_aggregate` functions implemented elsewhere:

```python
def evaluate(
    test_samples: List[TestSample],
    ground_truths: List[GroundTruth],
    inferences: List[Inference],
    test_cases: TestCases,
    # configuration: EvaluatorConfiguration,  # uncomment when configuration is used
) -> EvaluationResults:
    # compute per-sample metrics for each test sample
    per_sample_metrics = [compute_per_sample(gt, inf) for gt, inf in zip(ground_truths, inferences)]

    # compute aggregate metrics across all test cases using `test_cases.iter(...)`
    aggregate_metrics: List[Tuple[TestCase, MetricsTestCase]] = []
    for test_case, *s in test_cases.iter(test_samples, ground_truths, inferences, per_sample_metrics):
        # subset of `test_samples`/`ground_truths`/`inferences`/`test_sample_metrics` in given test case
        tc_test_samples, tc_ground_truths, tc_inferences, tc_per_sample_metrics = s
        aggregate_metrics.append((test_case, compute_aggregate(tc_per_sample_metrics)))

    # if desired, compute and add `plots_test_case` and `metrics_test_suite`
    return EvaluationResults(
        metrics_test_sample=list(zip(test_samples, per_sample_metrics)),
        metrics_test_case=aggregate_metrics,
    )
```

The control flow is in general more streamlined than with [`Evaluator`][kolena.workflow.Evaluator], but requires a
couple of assumptions to hold:

- Test-sample-level metrics do not vary by test case
- Ground truths corresponding to a given test sample do not vary by test case

This `BasicEvaluatorFunction` is provided to the test run at runtime, and is expected to have the following signature:

:param List[TestSample] test_samples: A list of distinct [`TestSample`][kolena.workflow.TestSample] values
    that correspond to all test samples in the test run.
:param List[GroundTruth] ground_truths: A list of [`GroundTruth`][kolena.workflow.GroundTruth] values
    corresponding to and sequenced in the same order as `test_samples`.
:param List[Inference] inferences: A list of [`Inference`][kolena.workflow.Inference] values corresponding
    to and sequenced in the same order as `test_samples`.
:param TestCases test_cases: An instance of [`TestCases`][kolena.workflow.TestCases], used to provide iteration
    groupings for evaluating test-case-level metrics.
:param EvaluatorConfiguration evaluator_configuration: The
    [`EvaluatorConfiguration`][kolena.workflow.EvaluatorConfiguration] to use when performing the evaluation. This
    parameter may be omitted in the function definition for implementations that do not use any configuration object.
:rtype: EvaluationResults
:return: An [`EvaluationResults`][kolena.workflow.EvaluationResults] object tracking the test-sample-level,
    test-case-level and test-suite-level metrics and plots for the input collection of test samples.
"""


class _TestCases(TestCases):
    def __init__(
        self,
        test_case_membership: List[Tuple[TestCase, List[TestSample]]],
        test_run_id: int,
        n_configurations: int,
    ):
        self._test_case_membership = test_case_membership
        self._test_run_id = test_run_id
        self._wip_configuration: Optional[EvaluatorConfiguration] = None
        self._n_test_cases_and_configurations = max(n_configurations, 1) * len(self._test_case_membership)
        self._n_test_cases_processed = 0

    def iter(
        self,
        test_samples: List[TestSample],
        ground_truths: List[GroundTruth],
        inferences: List[Inference],
        metrics_test_sample: List[MetricsTestSample],
    ) -> Iterator[Tuple[TestCase, List[TestSample], List[GroundTruth], List[Inference], List[MetricsTestSample]]]:
        metrics_by_test_sample = {
            self._test_sample_key(ts): (ts, gt, inf, mts)
            for ts, gt, inf, mts in zip(test_samples, ground_truths, inferences, metrics_test_sample)
        }
        for tc, test_case_test_samples in self._test_case_membership:
            samples, gts, infs, metrics = [], [], [], []
            for ts in test_case_test_samples:
                test_sample_key = self._test_sample_key(ts)
                _, gt, inf, mts = metrics_by_test_sample[test_sample_key]
                samples.append(ts)
                gts.append(gt)
                infs.append(inf)
                metrics.append(mts)
            yield tc, samples, gts, infs, metrics
            self._n_test_cases_processed += 1
            self._update_progress(tc)

    def _update_progress(self, test_case: TestCase) -> None:
        if not is_client_initialized():
            return

        config_description = (
            f" {_configuration_description(self._wip_configuration)}" if self._wip_configuration else ""
        )
        message = f"computed metrics for test case '{test_case.name}' (v{test_case.version}){config_description}"
        progress = self._n_test_cases_processed / self._n_test_cases_and_configurations

        log.info(message)
        request = API.UpdateMetricsStatusRequest(
            test_run_id=self._test_run_id,
            progress=progress,
            message=message,
        )
        res = krequests.put(
            endpoint_path=API.Path.UPDATE_METRICS_STATUS.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(res)

    def _set_configuration(self, configuration: Optional[EvaluatorConfiguration]) -> None:
        self._wip_configuration = configuration
        config_description = (
            f" {_configuration_description(self._wip_configuration)}" if self._wip_configuration else ""
        )
        log.info(f"computing metrics{config_description}")

    @staticmethod
    def _test_sample_key(ts: TestSample) -> str:
        return json.dumps(ts._to_dict(), sort_keys=True)


def _is_configured(evaluator: BasicEvaluatorFunction) -> bool:
    param_values = list(signature(evaluator).parameters.values())
    return len(param_values) == 5 and issubclass(param_values[4].annotation, EvaluatorConfiguration)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
import time
from abc import ABCMeta
from collections import defaultdict
from contextlib import contextmanager
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import pandas as pd
from pydantic import validate_arguments
from pydantic.dataclasses import dataclass

from kolena._api.v1.core import BulkProcessStatus
from kolena._api.v1.core import TestCase as CoreAPI
from kolena._api.v1.generic import TestCase as API
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame
from kolena._utils.consts import BatchSize
from kolena._utils.consts import FieldName
from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.frozen import Frozen
from kolena._utils.instrumentation import telemetry
from kolena._utils.instrumentation import WithTelemetry
from kolena._utils.serde import from_dict
from kolena._utils.validators import validate_name
from kolena._utils.validators import ValidatorConfig
from kolena.errors import IncorrectUsageError
from kolena.errors import NotFoundError
from kolena.workflow import GroundTruth
from kolena.workflow import TestSample
from kolena.workflow._datatypes import TestCaseEditorDataFrame
from kolena.workflow._datatypes import TestSampleDataFrame
from kolena.workflow._validators import assert_workflows_match
from kolena.workflow.test_sample import _METADATA_KEY
from kolena.workflow.workflow import Workflow


def _to_editor_data_frame(
    edits: List[Tuple[TestSample, Optional[GroundTruth], bool]],
    test_case_name: Optional[str] = None,
) -> TestCaseEditorDataFrame:
    records = [
        (
            test_case_name,
            test_sample._data_type().value,
            test_sample._to_dict(),
            test_sample._to_metadata_dict(),
            ground_truth._to_dict() if ground_truth is not None else None,
            remove,
        )
        for test_sample, ground_truth, remove, in edits
    ]
    columns = [
        "test_case_name",
        "test_sample_type",
        "test_sample",
        "test_sample_metadata",
        "ground_truth",
        "remove",
    ]
    df = pd.DataFrame(records, columns=columns)
    return TestCaseEditorDataFrame(validate_df_schema(df, TestCaseEditorDataFrame.get_schema(), trusted=True))


class TestCase(Frozen, WithTelemetry, metaclass=ABCMeta):
    """
    A test case holds a list of [test samples][kolena.workflow.TestSample] paired with
    [ground truths][kolena.workflow.GroundTruth] representing a testing dataset or a slice of a testing dataset.

    Rather than importing this class directly, use the `TestCase` type definition returned from
    [`define_workflow`][kolena.workflow.define_workflow.define_workflow].
    """

    workflow: Workflow
    """
    The workflow of this test case. Automatically populated when constructing via test case type returned from
    [`define_workflow`][kolena.workflow.define_workflow.define_workflow].
    """

    name: str
    """The unique name of this test case. Cannot be changed after creation."""

    version: int
    """The version of this test case. A test case's version is automatically incremented whenever it is edited via
    [`TestCase.edit`][kolena.workflow.TestCase.edit]."""

    description: str
    """Free-form, human-readable description of this test case. Can be edited at any time via
    [`TestCase.edit`][kolena.workflow.TestCase.edit]."""

    _id: int

    @telemetry
    def __init_subclass__(cls) -> None:
        if not hasattr(cls, "workflow"):
            raise NotImplementedError(f"{cls.__name__} must implement class attribute 'workflow'")
        super().__init_subclass__()

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        name: str,
        version: Optional[int] = None,
        description: Optional[str] = None,
        test_samples: Optional[List[Tuple[TestSample, GroundTruth]]] = None,
        reset: bool = False,
    ):
        if type(self) == TestCase:
            raise Exception("<TestCase> must be subclassed.")

        validate_name(name, FieldName.TEST_CASE_NAME)
        self._validate_test_samples(test_samples)

        try:
            self._populate_from_other(self.load(name, version))
            if description is not None and self.description != description and not reset:
                log.warn("test case already exists, not updating description when reset=False")
            if test_samples is not None:
                if self.version > 0 and not reset:
                    log.warn("not updating test samples for test case that has already been edited when reset=False")
                else:
                    self._hydrate(test_samples, description)
        except NotFoundError:
            self._populate_from_other(self.create(name, description, test_samples))
        self._freeze()

    @classmethod
    def _validate_test_samples(cls, test_samples: Optional[List[Tuple[TestSample, GroundTruth]]] = None) -> None:
        if not test_samples:
            return

        test_sample_type = cls.workflow.test_sample_type
        ground_truth_type = cls.workflow.ground_truth_type
        if any(
            not (isinstance(sample, test_sample_type) and isinstance(gt, ground_truth_type))
            for sample, gt in test_samples
        ):
            raise TypeError(f"test_samples should be list of tuple of ({test_sample_type}, {ground_truth_type})")

    def _populate_from_other(self, other: "TestCase") -> None:
        with self._unfrozen():
            self._id = other._id
            self.name = other.name
            self.version = other.version
            self.description = other.description
            self.workflow = other.workflow

    @classmethod
    def _create_from_data(cls, data: CoreAPI.EntityData) -> "TestCase":
        assert_workflows_match(cls.workflow.name, data.workflow)
        obj = cls.__new__(cls)
        obj._id = data.id
        obj.name = data.name
        obj.workflow = cls.workflow
        obj.version = data.version
        obj.description = data.description
        obj._freeze()
        return obj

    def _hydrate(self, test_samples: List[Tuple[TestSample, GroundTruth]], description: Optional[str] = None) -> None:
        if len(test_samples) == 0:
            log.warn("no test samples provided, unable to populate test case")
            return
        with self.edit(reset=True) as editor:
            if description is not None:
                editor.description(description)
            for test_sample, ground_truth in test_samples:
                editor.add(test_sample, ground_truth)

    @classmethod
    def create(
        cls,
        name: str,
        description: Optional[str] = None,
        test_samples: Optional[List[Tuple[TestSample, GroundTruth]]] = None,
    ) -> "TestCase":
        """
        Create a new test case with the provided name.

        :param name: The name of the new test case to create.
        :param description: Optional free-form description of the test case to create.
        :param test_samples: Optionally specify a set of test samples and ground truths to populate the test case.
        :return: The newly created test case.
        """
        cls._validate_test_samples(test_samples)
        validate_name(name, FieldName.TEST_CASE_NAME)
        request = CoreAPI.CreateRequest(name=name, description=description or "", workflow=cls.workflow.name)
        res = krequests.post(endpoint_path=API.Path.CREATE.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        data = from_dict(data_class=CoreAPI.EntityData, data=res.json())
        obj = cls._create_from_data(data)
        log.info(f"created test case '{name}' (v{obj.version})")
        if test_samples is not None:
            obj._hydrate(test_samples)
        return obj

    @classmethod
    def load(cls, name: str, version: Optional[int] = None) -> "TestCase":
        """
        Load an existing test case with the provided name.

        :param name: The name of the test case to load.
        :param version: Optionally specify a particular version of the test case to load. Defaults to the latest version
            when unset.
        :return: The loaded test case.
        """
        request = CoreAPI.LoadByNameRequest(name=name, version=version)
        res = krequests.put(endpoint_path=API.Path.LOAD.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        data = from_dict(data_class=CoreAPI.EntityData, data=res.json())
        log.info(f"loaded test case '{name}' (v{data.version})")
        return cls._create_from_data(data)

    def load_test_samples(self) -> List[Tuple[TestSample, GroundTruth]]:
        """
        Load all [`TestSample`s][kolena.workflow.TestSample] and [`GroundTruth`s][kolena.workflow.GroundTruth] contained
        in this test case.

        :return: A list of each test sample, paired with its ground truth, in this test case.
        """
        return list(self.iter_test_samples())

    def iter_test_samples(self) -> Iterator[Tuple[TestSample, GroundTruth]]:
        """
        Iterate through all [`TestSample`s][kolena.workflow.TestSample] and
        [`GroundTruth`s][kolena.workflow.GroundTruth] contained in this test case.

        :return: An iterator yielding each test sample, paired with its ground truth, in this test case.
        """
        log.info(f"loading test samples in test case '{self.name}' (v{self.version})")
        test_sample_type = self.workflow.test_sample_type
        ground_truth_type = self.workflow.ground_truth_type
        init_request = CoreAPI.InitLoadContentsRequest(batch_size=BatchSize.LOAD_SAMPLES.value, test_case_id=self._id)
        for df in _BatchedLoader.iter_data(
            init_request=init_request,
            endpoint_path=API.Path.INIT_LOAD_TEST_SAMPLES.value,
            df_class=TestSampleDataFrame,
        ):
            has_metadata = "test_sample_metadata" in df.columns
            for record in df.itertuples():
                metadata_field = record.test_sample_metadata if has_metadata else {}
                test_sample = test_sample_type._from_dict({**record.test_sample, _METADATA_KEY: metadata_field})
                ground_truth = ground_truth_type._from_dict(record.ground_truth)
                yield test_sample, ground_truth
        log.info(f"loaded test samples in test case '{self.name}' (v{self.version})")

    class Editor:
        @dataclass(frozen=True)
        class _Edit:
            test_sample: TestSample
            ground_truth: Optional[GroundTruth] = None
            remove: bool = False

        _edits: List[_Edit]
        _reset: bool
        _description: str
        _initial_description: str

        @validate_arguments(config=ValidatorConfig)
        def __init__(self, description: str, reset: bool) -> None:
            self._edits = []
            self._reset = reset
            self._description = description
            self._initial_description = description

        @validate_arguments(config=ValidatorConfig)
        def description(self, description: str) -> None:
            """Update the description of the test case."""
            self._description = description

        @validate_arguments(config=ValidatorConfig)
        def add(self, test_sample: TestSample, ground_truth: GroundTruth) -> None:
            """
            Add a test sample to the test case. When the test sample already exists in the test case, its ground truth
            is overwritten with the ground truth provided here.

            :param test_sample: The test sample to add.
            :param ground_truth: The ground truth for the test sample.
            """
            self._edits.append(self._Edit(test_sample, ground_truth=ground_truth))

        @validate_arguments(config=ValidatorConfig)
        def remove(self, test_sample: TestSample) -> None:
            """
            Remove a test sample from the test case. Does nothing if the test sample is not in the test case.

            :param test_sample: The test sample to remove.
            """
            self._edits.append(self._Edit(test_sample, remove=True))

        def _edited(self) -> bool:
            return len(self._edits) > 0 or self._reset or self._description != self._initial_description

        def _to_data_frame(self, test_case_name: Optional[str] = None) -> TestCaseEditorDataFrame:
            return _to_editor_data_frame(
                [(edit.test_sample, edit.ground_truth, edit.remove) for edit in self._edits],
                test_case_name,
            )

    @contextmanager
    def edit(self, reset: bool = False) -> Iterator[Editor]:
        """
        Edit this test case in a context:

        ```python
        with test_case.edit() as editor:
            # perform as many editing actions as desired
            editor.add(...)
            editor.remove(...)
        ```

        Changes are committed to the Kolena platform when the context is exited.

        :param reset: Clear any and all test samples currently in the test case.
        """
        editor = self.Editor(self.description, reset)

        yield editor

        if not editor._edited():
            return

        log.info(f"editing test case '{self.name}' (v{self.version})")
        init_response = init_upload()
        df_serialized = editor._to_data_frame().as_serializable()
        upload_data_frame(df=df_serialized, batch_size=BatchSize.UPLOAD_RECORDS.value, load_uuid=init_response.uuid)

        request = CoreAPI.CompleteEditRequest(
            test_case_id=self._id,
            current_version=self.version,
            description=editor._description,
            reset=editor._reset,
            uuid=init_response.uuid,
        )
        complete_res = krequests.put(
            endpoint_path=API.Path.COMPLETE_EDIT.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(complete_res)
        test_case_data = from_dict(data_class=CoreAPI.EntityData, data=complete_res.json())
        self._populate_from_other(self._create_from_data(test_case_data))
        log.success(f"edited test case '{self.name}' (v{self.version})")

    @classmethod
    def init_many(
        cls,
        data: List[Tuple[str, List[Tuple[TestSample, GroundTruth]]]],
        reset: bool = False,
    ) -> List["TestCase"]:
        """
        !!! note "Experimental"

            This function is considered **experimental**, so beware that it is subject to changes
            even in patch releases.

        Create, load or edit multiple test cases.

        ```python
        test_cases = TestCase.init_many([
            ("test-case 1", [(test_sample_1, ground_truth_1), ...]),
            ("test-case 2", [(test_sample_2, ground_truth_2), ...])
        ])

        test_suite = TestSuite("my test suite", test_cases=test_cases)
        ```

        Changes are committed to the Kolena platform together. If there is an error, none of the edits would take
        effect.

        :param data: A list of tuples where each tuple is a test case name and a set of test samples and ground truths
                     tuples for the test case.
        :param reset: If a test case of the same name already exists, overwrite with the provided test_samples.
        :return: The test cases.


        """

        if not hasattr(cls, "workflow"):
            raise NotImplementedError(f"{cls.__name__} must implement class attribute 'workflow'")

        if len({name for name, _ in data}) != len(data):
            raise IncorrectUsageError("Multiple edits to the same test case")

        for _, test_samples in data:
            if test_samples:
                cls._validate_test_samples(test_samples)

        cls._validate_test_samples()

        log.info(f"initializing {len(data)} test cases")
        start = time.time()

        df_test_samples = [
            _to_editor_data_frame(
                [(sample, ground_truth, False) for sample, ground_truth in test_samples],
                name,
            ).as_serializable()
            for name, test_samples in data
            if test_samples
        ]
        df_serialized = pd.concat(df_test_samples) if df_test_samples else pd.DataFrame()
        if len(df_serialized):
            load_uuid = init_upload().uuid
            upload_data_frame(df=df_serialized, batch_size=BatchSize.UPLOAD_RECORDS.value, load_uuid=load_uuid)
        else:
            load_uuid = None

        request = CoreAPI.BulkProcessRequest(
            test_cases=[CoreAPI.SingleProcessRequest(name=name, reset=reset) for name, _ in data],
            workflow=cls.workflow.name,
            uuid=load_uuid,
        )
        response = krequests.post(
            endpoint_path=API.Path.BULK_PROCESS.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(response)
        bulk_response = from_dict(data_class=CoreAPI.BulkProcessResponse, data=response.json())

        test_cases = []
        statuses = defaultdict(int)
        for test_case_data in bulk_response.test_cases:
            test_cases.append(cls._create_from_data(test_case_data.data))
            statuses[test_case_data.status] += 1

        end = time.time()
        status_msg = ", ".join([f"{s.value} {statuses[s]}" for s in BulkProcessStatus])
        log.info(f"initialized {len(data)} test cases: {status_msg} in {end - start:0.3f} seconds")

        return test_cases



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
from typing import Optional
from typing import Union

from deprecation import deprecated
from pydantic import Field
from pydantic.dataclasses import dataclass

from kolena.workflow import define_workflow
from kolena.workflow import EvaluatorConfiguration
from kolena.workflow import GroundTruth as BaseGroundTruth
from kolena.workflow import Image
from kolena.workflow import Inference as BaseInference
from kolena.workflow import Metadata
from kolena.workflow import MetricsTestCase
from kolena.workflow import MetricsTestSample
from kolena.workflow import MetricsTestSuite
from kolena.workflow.annotation import ClassificationLabel
from kolena.workflow.annotation import ScoredClassificationLabel


@dataclass(frozen=True)
class TestSample(Image):
    metadata: Metadata = Field(default_factory=dict)


@dataclass(frozen=True)
class GroundTruth(BaseGroundTruth):
    classification: ClassificationLabel


# DEPRECATED: preserved for compatibility only. Please use ScoredClassificationLabel instead
@dataclass(frozen=True)
class InferenceLabel(ClassificationLabel):
    confidence: float

    @deprecated(details="use `kolena.workflow.annotation.ScoredClassificationLabel`", deprecated_in="0.70.0")
    def __post_init__(self) -> None:
        ...

    @property
    def score(self) -> float:
        return self.confidence


@dataclass(frozen=True)
class Inference(BaseInference):
    inferences: List[Union[ScoredClassificationLabel, InferenceLabel]]


_workflow, TestCase, TestSuite, Model = define_workflow("Multiclass Classification", TestSample, GroundTruth, Inference)


@dataclass(frozen=True)
class TestSampleMetrics(MetricsTestSample):
    classification: Optional[Union[ScoredClassificationLabel, InferenceLabel]]
    margin: Optional[float]
    is_correct: bool


@dataclass(frozen=True)
class AggregatedMetrics:
    F1: float
    Precision: float
    Recall: float
    FPR: float


@dataclass(frozen=True)
class TestCaseMetrics(MetricsTestCase):
    n_labels: int
    n_correct: int
    n_incorrect: int
    accuracy: float
    macro_precision: float
    macro_recall: float
    macro_f1: float
    macro_tpr: float
    macro_fpr: float


@dataclass(frozen=True)
class TestSuiteMetrics(MetricsTestSuite):
    n_images: int
    n_correct: int
    mean_test_case_accuracy: float


@dataclass(frozen=True)
class ThresholdConfiguration(EvaluatorConfiguration):
    threshold: Optional[float] = None

    def display_name(self) -> str:
        if self.threshold:
            return f"Confidence Above Threshold (threshold={self.threshold})"
        return "Max Confidence"



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from pydantic import validate_arguments

from kolena import detection
from kolena._utils.validators import ValidatorConfig
from kolena.detection._internal.metadata import _from_dict
from kolena.detection._internal.metadata import MetadataElement
from kolena.detection._internal.test_image import BaseTestImage
from kolena.detection.ground_truth import ClassificationLabel
from kolena.detection.ground_truth import GroundTruth


class TestImage(BaseTestImage):
    """An image with associated ground truth labels for testing."""

    locator: str
    """Pointer to the bucket location of this image, e.g. `gs://my-bucket/my-dataset/example.png`."""

    dataset: str
    """The source dataset this image belongs to."""

    labels: List[str]
    """
    Zero or more ground truth labels for this image. For binary classifiers, an arbitrary string such as `positive`
    may be used. Not surfaced during testing.
    """

    metadata: Dict[str, MetadataElement]
    """
    Arbitrary metadata associated with this image. This metadata is surfaced during testing and may be used as model
    inputs as necessary.

    Certain metadata values can be visualized in the web platform when viewing results:

    - [`Annotation`][kolena.classification.metadata.Annotation] objects are overlaid on the main image
    - [`Asset`][kolena.classification.metadata.Asset] objects containing locators pointing to images, e.g.
        ``gs://my-bucket/my-dataset/example-1channel.png``, are displayed

    See the [metadata documentation][kolena.classification.metadata.Annotation] for more details.
    """

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        locator: str,
        dataset: Optional[str] = None,
        labels: Optional[List[str]] = None,
        metadata: Optional[Dict[str, MetadataElement]] = None,
    ):
        super().__init__(locator, dataset, metadata)
        self.labels = labels or []

    def filter(self, predicate: Callable[[str], bool]) -> "TestImage":
        """
        Return a copy of this test image with ground truth labels filtered to only those that match the provided
        predicate.

        :param predicate: Function accepting a string label and returning a boolean indicating whether or not to include
            the ground truth label.
        :return: A new test image with labels filtered by the predicate.
        """
        return TestImage(**{**self._fields(), "labels": [label for label in self.labels if predicate(label)]})

    @classmethod
    def _meta_keys(cls) -> List[str]:
        return detection.TestImage._meta_keys()

    @classmethod
    def _from_record(cls, record: Any) -> "TestImage":
        return TestImage(
            record.locator,
            dataset=record.dataset,
            labels=[GroundTruth._from_dict(gt_dict).label for gt_dict in record.ground_truths or []],
            metadata=_from_dict(record.metadata),
        )

    @classmethod
    def _to_record(cls, image: "TestImage") -> Tuple[str, Optional[str], List[Dict[str, Any]], Dict[str, Any]]:
        return detection.TestImage._to_record(image._to_detection())

    def _to_detection(self) -> detection.TestImage:
        return detection.TestImage(
            locator=self.locator,
            dataset=self.dataset,
            ground_truths=[ClassificationLabel(label) for label in self.labels],
            metadata=self.metadata,
        )

    # TODO: remove implementation in favor of Frozen.__eq__ once label ordering is ensured upstream
    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and {**self.__dict__, "labels": sorted(self.labels)} == {
            **other.__dict__,
            "labels": sorted(other.labels),
        }



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np

from kolena._utils import log
from kolena.classification.multiclass import InferenceLabel
from kolena.workflow.annotation import ScoredClassificationLabel


def get_label_confidence(label: str, inference_labels: List[Union[ScoredClassificationLabel, InferenceLabel]]) -> float:
    for inf_label in inference_labels:
        if inf_label.label == label:
            return inf_label.score
    return 0


def roc_curve(y_true: List[int], y_score: List[float]) -> Tuple[List[float], List[float]]:
    # Convert inputs to numpy arrays
    y_true = np.array(y_true)
    y_score = np.array(y_score)

    # Sort the predictions by descending order of confidence
    sorted_indices = np.argsort(y_score)[::-1]
    y_score = y_score[sorted_indices]
    y_true = y_true[sorted_indices]
    distinct_value_indices = np.where(np.diff(y_score[sorted_indices]))[0]
    threshold_idxs = np.r_[distinct_value_indices, y_true.size - 1]

    # Compute the cumulative sums of true positives and false positives
    tps = np.cumsum(y_true)[threshold_idxs]
    fps = np.cumsum(1 - y_true)[threshold_idxs]

    # Drop collinear points (copied from sklearn)
    if len(fps) > 2:
        optimal_indices = np.where(np.r_[True, np.logical_or(np.diff(fps, 2), np.diff(tps, 2)), True])[0]
        fps = fps[optimal_indices]
        tps = tps[optimal_indices]

    fps = np.r_[0, fps]
    tps = np.r_[0, tps]

    # Map cumulative sums to tpr and fpr
    if fps[-1] <= 0:
        # No negative samples in y_true, false positive value should be meaningless
        fpr = []
    else:
        fpr = fps / fps[-1]
        fpr = fpr.tolist()
    if tps[-1] <= 0:
        # No positive samples in y_true, true positive value should be meaningless
        tpr = []
    else:
        tpr = tps / tps[-1]
        tpr = tpr.tolist()

    return fpr, tpr


def get_histogram_range(scores: List[float]) -> Optional[Tuple[float, float, int]]:  # min, max, num_buckets
    min_score, max_score = min(scores), max(scores)
    if min_score < 0.0 or max_score > 1.0:
        log.warn(
            f"scores out of range for confidence histograms: expecting [0, 1], got [{min_score:.3f}, {max_score:.3f}]",
        )
        return None

    bin_range_options = np.linspace(0, 1, 51)
    min_range, max_range, bucket_fenceposts = 0, 1, 0
    for option in bin_range_options:
        if min_score >= option > min_range:
            min_range = option
        if max_score <= option < max_range:
            max_range = option
    for option in bin_range_options:
        if min_range <= option <= max_range:
            bucket_fenceposts += 1
    return min_range, max_range, bucket_fenceposts - 1



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from pydantic.dataclasses import dataclass

from kolena._api.v1.batched_load import BatchedLoad


class TestImages:
    class Path(str, Enum):
        INIT_LOAD_REQUEST = "/fr/test-images/load/init"
        COMPLETE_REGISTER = "/fr/test-images/register/complete"

    @dataclass(frozen=True)
    class LoadRequest:
        include_augmented: bool
        data_source: Optional[str] = None
        test_case_id: Optional[int] = None
        test_suite_id: Optional[int] = None

        def __post_init__(self) -> None:
            if sum(1 for value in [self.data_source, self.test_case_id, self.test_suite_id] if value is not None) > 1:
                raise ValueError("must populate at most one of: data_source, test_case_id, test_suite_id")

    @dataclass(frozen=True)
    class InitLoadRequest(LoadRequest, BatchedLoad.BaseInitDownloadRequest):
        ...

    @dataclass(frozen=True)
    class RegisterResponse:
        n_images_inserted: int
        n_images_updated: int
        n_tags_inserted: int
        n_tags_updated: int


class TestCase:
    class Path(str, Enum):
        CREATE = "/fr/test-case/create"
        LOAD_BY_NAME = "/fr/test-case/load-by-name"
        COMPLETE_EDIT = "/fr/test-case/edit/complete"
        INIT_LOAD_DATA = "/fr/test-case/load-data/init"

    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        description: str

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str
        version: Optional[int] = None

    @dataclass(frozen=True)
    class LoadDataRequest:
        test_case_id: int

    @dataclass(frozen=True)
    class InitLoadDataRequest(LoadDataRequest, BatchedLoad.BaseInitDownloadRequest):
        ...

    @dataclass(frozen=True)
    class EditRequest:
        test_case_id: int
        current_version: int
        name: str  # TODO: deprecate and remove (unused)
        description: str

    @dataclass(frozen=True)
    class CompleteEditRequest(EditRequest, BatchedLoad.WithLoadUUID):
        ...

    @dataclass(frozen=True)
    class EntityData:
        id: int
        name: str
        version: int
        description: str
        image_count: int
        pair_count_genuine: int
        pair_count_imposter: int


class TestSuite:
    class Path(str, Enum):
        CREATE = "/fr/test-suite/create"
        LOAD_BY_NAME = "/fr/test-suite/load-by-name"
        EDIT = "/fr/test-suite/edit"
        DELETE = "/fr/test-suite/delete"

    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        description: str

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str
        version: Optional[int] = None

    @dataclass(frozen=True)
    class EditRequest:
        test_suite_id: int
        current_version: int
        name: str
        description: str
        baseline_test_case_ids: List[int]
        non_baseline_test_case_ids: List[int]

    @dataclass(frozen=True)
    class EntityData:
        id: int
        name: str
        version: int
        description: str
        baseline_test_cases: List[TestCase.EntityData]
        non_baseline_test_cases: List[TestCase.EntityData]
        baseline_image_count: int
        baseline_pair_count_genuine: int
        baseline_pair_count_imposter: int

    @dataclass(frozen=True)
    class DeleteRequest:
        test_suite_id: int


class Model:
    class Path(str, Enum):
        CREATE = "/fr/model/create"
        LOAD_BY_NAME = "/fr/model/load-by-name"
        INIT_LOAD_PAIR_RESULTS = "/fr/model/load-pair-results/init"
        DELETE = "/fr/model/delete"

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str

    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        metadata: Dict[str, Any]

    @dataclass(frozen=True)
    class LoadPairResultsRequest:
        model_id: int
        # exactly one of the two must be present
        test_suite_id: Optional[int] = None
        test_case_id: Optional[int] = None

        def __post_init_post_parse__(self) -> None:
            if self.test_suite_id is None and self.test_case_id is None:
                raise ValueError("either test_case_id or test_suite_id must be present")
            if self.test_suite_id is not None and self.test_case_id is not None:
                raise ValueError("only one of test_case_id or test_suite_id may be present")

    @dataclass(frozen=True)
    class InitLoadPairResultsRequest(LoadPairResultsRequest, BatchedLoad.BaseInitDownloadRequest):
        ...

    @dataclass(frozen=True)
    class DeleteRequest:
        model_id: int


class TestRun:
    class Path(str, Enum):
        CREATE_OR_RETRIEVE = "/fr/test-run/create-or-retrieve"
        INIT_LOAD_REMAINING_IMAGES = "/fr/test-run/load-remaining-images/init"
        COMPLETE_UPLOAD_IMAGE_RESULTS = "/fr/test-run/upload-image-results/complete"
        INIT_LOAD_REMAINING_PAIRS = "/fr/test-run/load-remaining-pairs/init"
        COMPLETE_UPLOAD_PAIR_RESULTS = "/fr/test-run/upload-pair-results/complete"
        MARK_CRASHED = "/fr/test-run/mark-crashed"

    @dataclass(frozen=True)
    class CreateOrRetrieveRequest:
        model_id: int
        test_suite_ids: List[int]
        reset: bool = False

    @dataclass(frozen=True)
    class MarkCrashedRequest:
        test_run_id: int

    @dataclass(frozen=True)
    class LoadRemainingImagesRequest:
        test_run_id: int
        batch_size: int
        load_all: bool = False

    @dataclass(frozen=True)
    class InitLoadRemainingImagesRequest(LoadRemainingImagesRequest, BatchedLoad.BaseInitDownloadRequest):
        ...

    @dataclass(frozen=True)
    class UploadImageResultsRequest(BatchedLoad.WithLoadUUID):
        test_run_id: int
        reset: bool = False

    @dataclass(frozen=True)
    class UploadImageResultsResponse:
        n_uploaded: int

    @dataclass(frozen=True)
    class LoadRemainingPairsRequest:
        test_run_id: int
        batch_size: int
        load_all: bool = False

    @dataclass(frozen=True)
    class InitLoadRemainingPairsRequest(LoadRemainingPairsRequest, BatchedLoad.BaseInitDownloadRequest):
        ...

    @dataclass(frozen=True)
    class InitLoadRemainingPairsPartialResponse:
        pairs: BatchedLoad.InitDownloadPartialResponse
        embeddings: BatchedLoad.InitDownloadPartialResponse

    @dataclass(frozen=True)
    class UploadPairResultsRequest(BatchedLoad.WithLoadUUID):
        test_run_id: int
        reset: bool = False

    @dataclass(frozen=True)
    class UploadPairResultsResponse:
        n_uploaded: int


class Asset:
    class Path(str, Enum):
        CONFIG = "/fr/asset/config"
        BULK_UPLOAD = "/fr/asset/upload/bulk"

    @dataclass(frozen=True)
    class Config:
        bucket: str
        prefix: str

    @dataclass(frozen=True)
    class BulkUploadResponse:
        n_uploaded: int



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

from kolena.workflow.annotation import ScoredLabeledBoundingBox

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import numpy as np

from kolena._extras.metrics.sklearn import sklearn_metrics
from kolena._utils import log
from kolena.workflow import ConfusionMatrix
from kolena.workflow import Curve
from kolena.workflow import CurvePlot
from kolena.workflow.metrics import InferenceMatches
from kolena.workflow.metrics import MulticlassInferenceMatches


def filter_inferences(
    inferences: List[ScoredLabeledBoundingBox],
    confidence_score: Optional[float] = None,
    labels: Optional[Set[str]] = None,
) -> List[ScoredLabeledBoundingBox]:
    filtered_by_confidence = (
        [inf for inf in inferences if inf.score >= confidence_score] if confidence_score else inferences
    )
    if labels is None:
        return filtered_by_confidence
    return [inf for inf in filtered_by_confidence if inf.label in labels]


def _compute_sklearn_arrays(
    all_matches: List[Union[MulticlassInferenceMatches, InferenceMatches]],
) -> Tuple[np.ndarray, np.ndarray]:
    y_true: List[int] = []
    y_score: List[float] = []
    for image_bbox_matches in all_matches:
        for _, bbox_inf in image_bbox_matches.matched:  # TP (if above threshold)
            y_true.append(1)
            y_score.append(bbox_inf.score)
        for _ in image_bbox_matches.unmatched_gt:  # FN
            y_true.append(1)
            y_score.append(-1)
        for bbox_inf in image_bbox_matches.unmatched_inf:  # FP (if above threshold)
            y_true.append(0)
            y_score.append(bbox_inf.score)
    return np.array(y_true), np.array(y_score)


def _compute_threshold_curve(
    y_true: np.ndarray,
    y_score: np.ndarray,
    curve_type: Literal["pr", "f1"],
    curve_label: Optional[str] = None,
) -> Optional[Curve]:
    if len(y_score) < 1:
        return None

    potential_thresholds = np.unique(y_score[y_score >= 0.0]).tolist()  # sorts

    if len(potential_thresholds) >= 501:
        potential_thresholds = list(np.linspace(min(potential_thresholds), max(potential_thresholds), 501))

    precisions: List[float] = []
    recalls: List[float] = []
    thresholds: List[float] = []
    f1s: List[float] = []
    for threshold in potential_thresholds:
        y_pred = [1 if score >= threshold else 0 for score in y_score]
        precision, recall, f1, _ = sklearn_metrics.precision_recall_fscore_support(
            y_true,
            y_pred,
            average="binary",
            zero_division=0,
        )

        # avoid curves with one x-value and two y-values
        if recall in recalls:
            idx = recalls.index(recall)
            precisions[idx] = max(precisions[idx], precision)
            # if an old point is replaced, replace the f1s and thresholds
            if precisions[idx] == precision:
                thresholds[idx] = threshold
                f1s[idx] = f1
        else:
            precisions.append(precision)
            recalls.append(recall)
            thresholds.append(threshold)
            f1s.append(f1)

    # omit curves with no points
    if len(f1s) == 0 or len(precisions) == 0 or len(recalls) == 0:
        return None

    if curve_type == "f1":
        return (
            Curve(x=thresholds, y=f1s, label=curve_label, extra=dict(Precision=precisions, Recall=recalls))
            if len(f1s) >= 2
            else None
        )

    # add a point to start the PR curve on the vertical axis if needed
    if 0.0 not in recalls:
        minpos = recalls.index(min(recalls))
        precisions.append(precisions[minpos])
        recalls.append(0.0)
        # maintain the same lengths for f1s and thresholds
        f1s.append(f1s[minpos])
        thresholds.append(thresholds[minpos])
    return (
        Curve(x=recalls, y=precisions, label=curve_label, extra=dict(F1=f1s, Threshold=thresholds))
        if len(recalls) >= 2
        else None
    )


def _compute_multiclass_curves(
    all_matches: List[MulticlassInferenceMatches],
    curve_type: Literal["pr", "f1"],
) -> Optional[List[Curve]]:
    curves: List[Curve] = []
    y_true_score_by_label = _compute_sklearn_arrays_by_class(all_matches)

    for label, (y_true, y_score) in sorted(y_true_score_by_label.items(), key=lambda x: x[0]):
        curve = _compute_threshold_curve(y_true, y_score, curve_type, label)
        if curve is not None:
            curves.append(curve)
    return curves if curves else None


def compute_pr_curve(
    all_matches: List[Union[MulticlassInferenceMatches, InferenceMatches]],
    curve_label: Optional[str] = None,
) -> Optional[Curve]:
    """
    Creates a PR (precision and recall) curve.

    :param all_matches: A list of multiclass or singleclass matching results.
    :param curve_label: The label of the curve.
    :return: :class:`Curve` for the PR curve.
    """
    y_true, y_score = _compute_sklearn_arrays(all_matches)
    return _compute_threshold_curve(y_true, y_score, "pr", curve_label)


def compute_pr_plot(
    all_matches: List[Union[MulticlassInferenceMatches, InferenceMatches]],
    curve_label: Optional[str] = None,
) -> Optional[CurvePlot]:
    """
    Creates a PR (precision and recall) plot.

    :param all_matches: A list of multiclass or singleclass matching results.
    :param curve_label: The label of the curve.
    :return: :class:`CurvePlot` for the PR curve.
    """
    curve = compute_pr_curve(all_matches, curve_label)

    return (
        CurvePlot(
            title="Precision vs. Recall",
            x_label="Recall",
            y_label="Precision",
            curves=[curve],
        )
        if curve
        else None
    )


def compute_pr_plot_multiclass(
    all_matches: List[MulticlassInferenceMatches],
) -> Optional[CurvePlot]:
    """
    Creates a PR (precision-recall) curve for the multiclass object detection workflow.
    For `n` labels, each plot has `n+1` curves. One for the test case, and one per label.

    :param all_matches: a list of multiclass matching results.
    :return: :class:`CurvePlot` for the PR curves of the test case and each label.
    """
    pr_curves: Optional[List[Curve]] = _compute_multiclass_curves(all_matches, "pr")

    return (
        CurvePlot(
            title="Precision vs. Recall Per Class",
            x_label="Recall",
            y_label="Precision",
            curves=pr_curves,
        )
        if pr_curves
        else None
    )


def compute_f1_plot(
    all_matches: List[Union[MulticlassInferenceMatches, InferenceMatches]],
    curve_label: Optional[str] = None,
) -> Optional[CurvePlot]:
    """
    Creates a F1-threshold (confidence threshold) plot.

    :param all_matches: A list of multiclass or singleclass matching results.
    :param curve_label: The label of the curve.
    :return: :class:`CurvePlot` for the F1-threshold curve.
    """
    y_true, y_score = _compute_sklearn_arrays(all_matches)
    curve = _compute_threshold_curve(y_true, y_score, "f1", curve_label)

    return (
        CurvePlot(
            title="F1-Score vs. Confidence Threshold",
            x_label="Confidence Threshold",
            y_label="F1-Score",
            curves=[curve],
        )
        if curve
        else None
    )


def compute_f1_plot_multiclass(
    all_matches: List[MulticlassInferenceMatches],
) -> Optional[CurvePlot]:
    """
    Creates a F1-threshold (confidence threshold) curve for the multiclass object detection workflow.
    For `n` labels, each plot has `n+1` curves. One for the test case, and one per label.

    :param all_matches: a list of multiclass matching results.
    :param curve_label: the label of the main curve.
    :return: :class:`CurvePlot` for the F1-threshold curves of the test case and each label.
    """
    f1_curves: Optional[List[Curve]] = _compute_multiclass_curves(all_matches, "f1")

    return (
        CurvePlot(
            title="F1-Score vs. Confidence Threshold Per Class",
            x_label="Confidence Threshold",
            y_label="F1-Score",
            curves=f1_curves,
        )
        if f1_curves
        else None
    )


def compute_confusion_matrix_plot(
    all_matches: List[MulticlassInferenceMatches],
    plot_title: str = "Confusion Matrix",
) -> Optional[ConfusionMatrix]:
    """
    Creates a [`ConfusionMatrix`][kolena.workflow.ConfusionMatrix] for a multiclass workflow.

    :param all_matches: A list of multiclass matching results.
    :param plot_title: The title for the [`ConfusionMatrix`][kolena.workflow.ConfusionMatrix].
    :return: [`ConfusionMatrix`][kolena.workflow.ConfusionMatrix] with all actual and predicted labels, if there is more
        than one label in the provided `all_matches`.
    """
    labels: Set[str] = set()

    confusion_matrix: Dict[str, Dict[str, int]] = defaultdict(lambda: defaultdict(int))
    for match in all_matches:
        for gt, _ in match.matched:
            actual_label = gt.label
            confusion_matrix[actual_label][actual_label] += 1
            labels.add(actual_label)

        for gt, inf in match.unmatched_gt:
            actual_label = gt.label
            labels.add(actual_label)
            if inf is not None:
                predicted_label = inf.label
                confusion_matrix[actual_label][predicted_label] += 1
                labels.add(predicted_label)

    if len(labels) < 2:
        log.info(f"skipping confusion matrix for a single label: {labels}")
        return None

    ordered_labels = sorted(labels)
    matrix = []
    for actual_label in ordered_labels:
        matrix.append([confusion_matrix[actual_label][predicted_label] for predicted_label in ordered_labels])
    return ConfusionMatrix(title=plot_title, labels=ordered_labels, matrix=matrix)


def _compute_sklearn_arrays_by_class(
    all_matches: List[MulticlassInferenceMatches],
) -> Dict[str, Tuple[np.ndarray, np.ndarray]]:
    y_true_and_score_by_label: Dict[str, Tuple[np.ndarray, np.ndarray]] = {}

    labels: Set[str] = set()
    for match in all_matches:
        for _, bbox_inf in match.matched:
            labels.add(bbox_inf.label)
        for bbox_gt, _ in match.unmatched_gt:
            labels.add(bbox_gt.label)
        for bbox_inf in match.unmatched_inf:
            labels.add(bbox_inf.label)

    for label in labels:
        filtered_matchings = [
            InferenceMatches(
                matched=[(gt, inf) for gt, inf in match.matched if gt.label == label],
                unmatched_gt=[gt for gt, _ in match.unmatched_gt if gt.label == label],
                unmatched_inf=[inf for inf in match.unmatched_inf if inf.label == label],
            )
            for match in all_matches
        ]

        y_true, y_score = _compute_sklearn_arrays(filtered_matchings)
        y_true_and_score_by_label[label] = (y_true, y_score)

    return y_true_and_score_by_label


def _compute_optimal_f1_with_arrays(
    y_true: np.ndarray,
    y_score: np.ndarray,
) -> float:
    # optimal threshold is 0 if there is no relevant or true label
    if len(y_true) == 0 or np.all(y_true == 0):
        return 0.0

    precision, recall, thresholds = sklearn_metrics.precision_recall_curve(y_true, y_score)

    # delete last pr of (1,0)
    precision = precision[:-1]
    recall = recall[:-1]

    if thresholds[0] < 0:
        precision = precision[1:]
        recall = recall[1:]
        thresholds = thresholds[1:]
    if len(thresholds) == 0:
        return 0.0

    f1_scores = 2 * precision * recall / (precision + recall)
    f1_scores = np.nan_to_num(f1_scores, nan=0)
    max_f1_index = np.argmax(f1_scores)
    if f1_scores[max_f1_index] == 0:
        return 0.0
    else:
        return float(thresholds[max_f1_index])


def compute_optimal_f1_threshold(
    all_matches: List[Union[MulticlassInferenceMatches, InferenceMatches]],
) -> float:
    """
    Computes the optimal F1 threshold for matchings.

    :param all_matches: A list of matching results.
    :return: The optimal F1 threshold value, zero where invalid.
    """
    y_true, y_score = _compute_sklearn_arrays(all_matches)
    return _compute_optimal_f1_with_arrays(y_true, y_score)


def compute_optimal_f1_threshold_multiclass(
    all_matches: List[MulticlassInferenceMatches],
) -> Dict[str, float]:
    """
    Creates a mapping of label to optimal F1 thresholds for a multiclass workflow.

    :param all_matches: A list of multiclass matching results.
    :return: A dictionary with each label and its optimal F1 threshold value, zero where invalid.
    """
    optimal_thresholds: Dict[str, float] = {}

    y_true_score_by_label = _compute_sklearn_arrays_by_class(all_matches)
    for label, (y_true, y_score) in sorted(y_true_score_by_label.items(), key=lambda x: x[0]):
        optimal_thresholds[label] = _compute_optimal_f1_with_arrays(y_true, y_score)
    return optimal_thresholds


def compute_average_precision(precisions: List[float], recalls: List[float]) -> float:
    """
    Computes the average precision given a PR curve with the metrics methodology of PASCAL VOC.
    Based on the [PASCAL VOC code in Python](https://github.com/Cartucho/mAP).

    :param precisions: A list precision values from a PR curve.
    :param recalls: A list recall values from a PR curve.
    :return: The value of the average precision.
    """

    if len(precisions) != len(recalls):
        raise ValueError("precisions and recalls differ in length")

    if len(precisions) == 0:
        return 0

    pairs = sorted(zip(recalls, precisions), key=lambda x: x[0])
    recalls, precisions = zip(*pairs)

    # add (0,0) to left and (1,0) to right
    recalls = [0, *recalls, 1]
    precisions = [0, *precisions, 0]

    # make precisions monotonic decreasing
    for i in range(len(precisions) - 2, -1, -1):
        precisions[i] = max(precisions[i], precisions[i + 1])

    # indices where recall has changed
    recall_changed_indices = []
    for i in range(1, len(recalls)):
        if recalls[i] != recalls[i - 1]:
            recall_changed_indices.append(i)

    ap = 0.0
    for i in recall_changed_indices:
        ap += (recalls[i] - recalls[i - 1]) * precisions[i]
    return ap
